@@ -6,7 +6,7 @@ Mojolicious::Plugin::CGI - Run CGI script from Mojolicious
 
 =head1 VERSION
 
-0.23
+0.24
 
 =head1 DESCRIPTION
 
@@ -19,33 +19,60 @@ manner.
 =head2 Standard usage
 
   use Mojolicious::Lite;
+  plugin CGI => [ "/cgi-bin/script" => "/path/to/cgi/script.pl" ];
+
+Using the code above is enough to run C<script.pl> when accessing
+L<http://localhost:3000/cgi-bin/script>.
+
+=head2 Complex usage
 
-  plugin CGI => [ '/script' => '/path/to/cgi/script.pl' ];
   plugin CGI => {
-    route => '/mount/point',
-    script => '/path/to/cgi/script.pl',
+    # Specify the script and mount point
+    script => "/path/to/cgi/script.pl",
+    route  => "/some/route",
+
+    # %ENV variables visible from inside the CGI script
     env => {}, # default is \%ENV
-    errlog => '/path/to/file.log', # path to where STDERR from cgi script goes
-    before => sub { # called before setup and script start
+
+    # Path to where STDERR from cgi script goes
+    errlog => "/path/to/file.log",
+
+    # The "before" hook is called before script start
+    # It receives a Mojolicious::Controller which can be modified
+    before => sub {
       my $c = shift;
-      # modify QUERY_STRING
       $c->req->url->query->param(a => 123);
     },
   };
 
-  app->start;
+The above contains all the options you can pass on to the plugin.
+
+=head2 Running code refs
+
+  plugin CGI => {
+    route => "/some/path",
+    run   => sub {
+      my $cgi = CGI->new;
+      # ...
+    }
+  };
+
+Instead of calling a script, you can run a code block when accessing the route.
+This is (pretty much) safe, even if the code block modifies global state,
+since it runs in a separate fork/process.
 
 =head2 Support for semicolon in query string
 
-  plugin CGI => { support_semicolon_in_query_string => 1 };
+  plugin CGI => {
+    support_semicolon_in_query_string => 1,
+    ...
+  };
 
 The code above need to be added before other plugins or handler which use
 L<Mojo::Message::Request/url>. It will inject a C<before_dispatch>
 hook which saves the original QUERY_STRING, before it is split on
 "&" in L<Mojo::Parameters>.
 
-This is an EXPERIMENTAL feature.
-
 =cut
 
 use Mojo::Base 'Mojolicious::Plugin';
@@ -62,7 +89,7 @@ use constant DEBUG                => $ENV{MOJO_PLUGIN_CGI_DEBUG} || 0;
 use constant READ                 => 0;
 use constant WRITE                => 1;
 
-our $VERSION = '0.23';
+our $VERSION = '0.24';
 our %ORIGINAL_ENV = %ENV;
 
 =head1 ATTRIBUTES
@@ -93,10 +120,10 @@ by the CGI script.
 
 In addition to L</env>, these dynamic variables are set:
 
-  CONTENT_LENGTH, CONTENT_TYPE, HTTP_COOKIE, HTTP_HOST, HTTP_REFERER,
-  HTTP_USER_AGENT, HTTPS, PATH, PATH_INFO, QUERY_STRING, REMOTE_ADDR,
-  REMOTE_HOST, REMOTE_PORT, REMOTE_USER, REQUEST_METHOD, SCRIPT_NAME,
-  SERVER_PORT, SERVER_PROTOCOL.
+  CONTENT_LENGTH, CONTENT_TYPE, HTTP_COOKIE, HTTP_HOST, HTTP_IF_NONE_MATCH,
+  HTTP_REFERER, HTTP_USER_AGENT, HTTPS, PATH, PATH_INFO, QUERY_STRING,
+  REMOTE_ADDR, REMOTE_HOST, REMOTE_PORT, REMOTE_USER, REQUEST_METHOD,
+  SCRIPT_NAME, SERVER_PORT, SERVER_PROTOCOL.
 
 Additional static variables:
 
@@ -128,12 +155,13 @@ sub emulate_environment {
     %{$self->env},
     CONTENT_LENGTH => $content_length        || 0,
     CONTENT_TYPE   => $headers->content_type || '',
-    GATEWAY_INTERFACE => 'CGI/1.1',
-    HTTP_COOKIE       => $headers->cookie || '',
-    HTTP_HOST         => $headers->host || '',
-    HTTP_REFERER      => $headers->referrer || '',
-    HTTP_USER_AGENT   => $headers->user_agent || '',
-    HTTPS             => $req->is_secure ? 'YES' : 'NO',
+    GATEWAY_INTERFACE  => 'CGI/1.1',
+    HTTP_COOKIE        => $headers->cookie || '',
+    HTTP_HOST          => $headers->host || '',
+    HTTP_REFERER       => $headers->referrer || '',
+    HTTP_USER_AGENT    => $headers->user_agent || '',
+    HTTP_IF_NONE_MATCH => $headers->if_none_match || '',
+    HTTPS              => $req->is_secure ? 'YES' : 'NO',
 
     #PATH => $req->url->path,
     PATH_INFO => '/' . ($c->stash('path_info') || ''),
@@ -176,7 +204,11 @@ sub register {
     $self->{script} = shift @$args;
   }
   elsif ($args->{support_semicolon_in_query_string}) {
-    $app->hook(before_dispatch => sub { $_[0]->stash('cgi.query_string' => $_[0]->req->url->query->to_string); });
+    $app->hook(
+      before_dispatch => sub {
+        $_[0]->stash('cgi.query_string' => $_[0]->req->url->query->to_string);
+      }
+    );
     return;
   }
   else {
@@ -203,8 +235,6 @@ sub register {
       defined($pid = fork) or die "[CGI] Could not fork $name: $!";
 
       unless ($pid) {
-
-        # No need to Mojo::IOLoop->reset, since the CGI script is started with exec()
         my @STDERR = @stderr ? ('>&', fileno $stderr[WRITE]) : ('>>', $self->{errlog});
         warn "[CGI:$name:$$] <<< (@{[$stdin->slurp]})\n" if DEBUG;
         %ENV = $self->emulate_environment($c);
@@ -215,10 +245,17 @@ sub register {
         $| = 1;
         select STDOUT;
         $| = 1;
-        { exec $self->{script} }
-        die "Could not execute $self->{script}: $!";
-      }
 
+        if (my $code = $self->{run}) {
+          Mojo::IOLoop->reset;    # clean up
+          $code->($c);
+          exit;
+        }
+        else {
+          { exec $self->{script} }
+          die "Could not execute $self->{script}: $!";
+        }
+      }
       $log->debug("[CGI:$name:$pid] START $self->{script}");
       $pids->{$pid} = $name;
 
@@ -278,15 +315,15 @@ sub _stdout_cb {
     $buf .= $chunk;
     $buf =~ s/^(.*?\x0a\x0d?\x0a\x0d?)//s or return;    # false until all headers has been read from the CGI script
     $headers = $1;
-
     if ($headers =~ /^HTTP/) {
+      $c->res->code($1) if $headers =~ m!^HTTP (\d\d\d)!;    # borked CGI response if SERVER_PROTOCOL has no version
       $c->res->parse($headers);
     }
     else {
-      $c->res->code($headers =~ /Location:/ ? 302 : 200);
+      $c->res->code($1) if $headers =~ /^Status: (\d\d\d)/m;
+      $c->res->code($headers =~ /Location:/ ? 302 : 200) unless $c->res->code;
       $c->res->parse($c->res->get_start_line_chunk(0) . $headers);
     }
-
     $c->write($buf) if length $buf;
   };
 }
